#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
test_gls_unibox_api
----------------------------------

Tests for `gls_unibox_api` module.
"""
import os
import logging
from decimal import Decimal
from datetime import date, time

import pytest

from gls_unibox_api import fields
from gls_unibox_api.api import Response, Client, Shipment, logger

logger.setLevel(logging.DEBUG)
# create console handler and set level to debug
ch = logging.StreamHandler()
ch.setLevel(logging.DEBUG)
logger.addHandler(ch)

os_env = os.environ


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


class TestForm(object):
    f_char = fields.Char('f_char')
    f_int = fields.Int('f_int')
    f_long = fields.Long('f_long')
    f_decimal = fields.Decimal('f_decimal')
    f_date = fields.Date('f_date')
    f_date2 = fields.Date('f_date2', separator='.')
    f_time = fields.Time('f_time')


@pytest.fixture
def test_form():
    test_form = TestForm()
    test_form.values = {
        'f_char': 'hello world',
        'f_int': '1',
        'f_long': '1234567890987654321',
        'f_decimal': '12345,6',
        'f_date': '13121987',
        'f_date2': '13.12.1987',
        'f_time': '11:10',
    }
    return test_form


class TestFields:

    def test_char_field(self, test_form):
        assert test_form.f_char == 'hello world'

        # Set a new value and check
        new_value = 'hello'
        test_form.f_char = new_value
        assert test_form.f_char == new_value

    def test_int_field(self, test_form):
        assert test_form.f_int == 1

        # Set a new value and check
        new_value = 100
        test_form.f_int = new_value
        assert test_form.f_int == new_value

    def test_long_field(self, test_form):
        assert test_form.f_long == 1234567890987654321

        # Set a new value and check
        new_value = 1234567890123456
        test_form.f_long = new_value
        assert test_form.f_long == new_value

    def test_decimal_field(self, test_form):
        assert test_form.f_decimal == Decimal('12345.6')

        # Set a new value and check
        new_value = Decimal('12345.78')
        test_form.f_decimal = new_value
        assert test_form.f_decimal == new_value

    def test_date_field(self, test_form):
        assert test_form.f_date == date(1987, 12, 13)

        # Set a new value and check
        new_value = date(2009, 7, 1)
        test_form.f_date = new_value
        assert test_form.f_date == new_value

    def test_date2_field(self, test_form):
        assert test_form.f_date2 == date(1987, 12, 13)

        # Set a new value and check
        new_value = date(2009, 7, 1)
        test_form.f_date2 = new_value
        assert test_form.f_date2 == new_value

    def test_time_field(self, test_form):
        assert test_form.f_time == time(11, 10)

        # Set a new value and check
        new_value = time(10, 30)
        test_form.f_time = new_value
        assert test_form.f_time == new_value


@pytest.fixture
def client():
    return Client(
        os_env['GLS_SERVER'],
        os_env['GLS_PORT'],
        test=True
    )


@pytest.fixture
def shipment(client):
    shipment = Shipment(client)

    # Now set the values
    # The values are based on Standard Shipment example in docs 3.5.6
    shipment.software.name = 'Python'
    shipment.software.version = '2.7'
    shipment.printer_name = 'zebrazpl200'

    shipment.consignee.country = 'DE'
    shipment.consignee.zip = '99334'
    shipment.parcel_number = '463500000007'
    shipment.parcel_weight = Decimal('31.5')
    shipment.shipping_date = date.today()

    shipment.consignor.customer_number = 15082
    shipment.consignor.name = 'Fruchtzentrale Orkos'
    shipment.consignor.name2 = 'Grossmarkt Essen'
    shipment.consignor.name3 = 'Halle VI'
    shipment.consignor.street = 'Luetzowstr. 28a'
    shipment.consignor.country = 'DE'
    shipment.consignor.zip = '45141'
    shipment.consignor.place = 'Dortmund'
    shipment.consignor.label = 'Empfanger'
    shipment.consignor.consignor = 'Essen'

    shipment.consignee.customer_number_label = 'Kd-Nr'
    shipment.consignee.customer_number = 4600
    shipment.consignee.id_type = 'ID-Nr'
    shipment.consignee.id_value = 800018406

    shipment.consignee.salutation = 'Firma'
    shipment.consignee.name = 'GLS Germany'
    shipment.consignee.name2 = 'Depot 46'
    shipment.consignee.street = 'Huckarder Str. 91'
    shipment.consignee.place = 'Dortmund'
    shipment.consignee.zip = '44147'

    shipment.parcel = 1
    shipment.quantity = 2

    shipment.gls_contract = os_env['GLS_CONTRACT']
    shipment.gls_customer_id = os_env['GLS_CUSTOMER_ID']
    shipment.location = os_env['GLS_LOCATION']

    return shipment


class TestShipment:

    def test_shipment(self, shipment):
        # TODO: Based on the response parse and use it
        shipment.create_label()


class TestResponse:

    def test_parsing(self):
        response = Response.parse(
            read('example_response.txt')
        )
        assert response.consignee.name == u'GLS Germany'
        assert response.shipping_date == date(2009, 1, 26)
        assert response.values['zpl_content']



# -*- coding: utf-8 -*-
class TagMeta(type):
    """
    Metaclass for all Tags
    """
    def __new__(meta, name, bases, dct):
        if name != 'Tag':
            if 'code' not in dct:
                raise TypeError('code is not defined in tag: %s' % name)

            if 'description' not in dct:
                raise TypeError('description is not defined in tag: %s' % name)

            if 'max_length' in dct and \
                    not isinstance(dct['max_length'], (int, long)):
                raise TypeError(
                    'max_length: expected int or long. Got %s (%s)' % (
                        dct['max_length'], type(dct['max_length'])
                    )
                )

        return type.__new__(meta, name, bases, dct)


class Tag(object):
    """
    Base Tag related functionality
    """
    __metaclass__ = TagMeta

    def get_encoded_value(self):
        if not hasattr(self, 'value'):
            rv = self.code
        else:
            rv = '%s:%s' % (self.code, self.value)

        # Finally return the encoded value
        return rv.encode('ISO-8859-1')


class StartTag(Tag):
    "Start Tag of Datastream"
    code = '\\' * 5 + 'GLS' + '\\' * 5
    description = __doc__


class EndTag(Tag):
    "Start Tag of Datastream"
    code = '/' * 5 + 'GLS' + '/' * 5
    description = __doc__


class CancelParcel(Tag):
    "CANCELLATION for parcel number"
    code = 'T000'
    description = __doc__

    def __init__(self, parcel_number):
        self.value = parcel_number


class PrinterInterface(Tag):
    "Printer-interface, also for selection of the Default-interface"
    code = 'T020'
    description = __doc__



# -*- coding: utf-8 -*-
import inspect
import socket
import logging

logger = logging.getLogger('gls_unibox_api')


from . import fields
from .tags import StartTag, EndTag


class SoftwareInformation(fields.FieldGroup):
    "Software info"
    name = fields.Char('T050', length=20)
    version = fields.Char('T051', length=20)

    def __init__(self, *args, **kwargs):
        super(SoftwareInformation, self).__init__(*args, **kwargs)

        self.values = {}


class ServiceData(fields.FieldGroup):
    """
    Service Data Element
    """
    national_product = fields.Char('T200', length=3)
    # future fields.Char('T201', length=2)
    # future fields.Char('T202', length=2)
    cash_service = fields.Char('T203', length=2)
    express = fields.Char('T204', length=2)
    weight = fields.Char('T205', length=2)
    product_flag = fields.Char('T206', length=3)
    service_flag = fields.Char('T207', length=3)
    incoterms = fields.Int('T210', length=2)

    def __init__(self, *args, **kwargs):
        super(ServiceData, self).__init__(*args, **kwargs)

        self.values = {}


class PickUpAddress(fields.FieldGroup):
    """
    A pick up address Element
    """
    name = fields.Char('T900', length=50)
    name2 = fields.Char('T901', length=50)
    name3 = fields.Char('T902', length=50)
    street = fields.Char('T903', length=50)
    country = fields.Char('T904', length=2)
    zip = fields.Char('T905', length=10)
    place = fields.Char('T906', length=50)
    phone = fields.Char('T907', length=15)

    def __init__(self, *args, **kwargs):
        super(PickUpAddress, self).__init__(*args, **kwargs)

        self.values = {}


class PickUp(fields.FieldGroup):
    date = fields.Date('T908', separator='.')

    def __init__(self, *args, **kwargs):
        super(PickUp, self).__init__(*args, **kwargs)

        self.values = {}


class Consignor(fields.FieldGroup):
    "Consignor/Shipper"
    customer_number = fields.Long('T805', length=10)
    salutation = fields.Char('T809', length=50)
    name = fields.Char('T810', length=50)
    name2 = fields.Char('T811', length=50)
    name3 = fields.Char('T812', length=50)
    street = fields.Char('T820', length=50)
    country = fields.Char('T821', length=2)
    zip = fields.Char('T822', length=10)
    consignor = fields.Char('T823', length=50)
    place = fields.Char('T864', length=30)

    label = fields.Char('T850')

    def __init__(self, *args, **kwargs):
        super(Consignor, self).__init__(*args, **kwargs)

        self.values = {}


class Consignee(fields.FieldGroup):
    """
    A Structure representing an address
    """
    customer_number_label = fields.Char('T851', length=25)
    customer_number = fields.Long('T852', length=10)

    id_type = fields.Char('T853', length=25)
    id_value = fields.Char('T854', length=25)

    salutation = fields.Char('T859', length=50)
    name = fields.Char('T860', length=40)
    name2 = fields.Char('T861', length=40)
    name3 = fields.Char('T862', length=40)
    street = fields.Char('T863', length=50)
    place = fields.Char('T864', length=30)

    # Defined out of range
    zip = fields.Char('T330', length=10)
    country = fields.Char('T100', length=2)

    def __init__(self, *args, **kwargs):
        super(Consignee, self).__init__(*args, **kwargs)

        self.values = {}


class Client(object):
    """
    GLS Unibox connection API
    """

    def __init__(self, server, port, test=False):
        self.server = server
        self.port = int(port)
        self.test = test

    def get_socket_conn(self):
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.server, self.port))
        return s

    def request(self, tags):
        values = (StartTag.code + '|'.join(tags) + '|' + EndTag.code)
        logger.debug('Request::values::%s', values)

        s = self.get_socket_conn()
        s.sendall(values)

        response = ''
        while True:
            data = s.recv(1024)
            response += data
            if not data:
                break

        s.close()

        logger.debug('Response::values::%s', response)

        return response


class CommonMixin(object):
    """
    Mixin with fields common to all requests
    """
    software = SoftwareInformation()
    location = fields.Char('T8700', length=6)

    #: Special functions have the following possibilities
    #:
    #: * NOPRINT - Parcel data is returned through the stream
    #: * NOSAVE - For prechecking or testing, where neither the storing or
    #:   printing of the label happens.
    special_functions = fields.MultipleChoice('T090')
    printer_name = fields.Char('T021', length=25)


class ShipmentMixin(object):
    """
    Mixin with fields specific to the shipment
    """
    parcel_number = fields.Char('T400')

    epl_depot_number = fields.Char('T101')
    hub = fields.Char('T110')
    bar_code = fields.Long('T300')
    sorting_flag = fields.Int('T310')

    last_routing_update = fields.Date('T520')
    parcel_weight = fields.Decimal('T530')
    shipping_date = fields.Date('T545', separator='.')

    consignee = Consignee()
    consignor = Consignor()
    pick_up_address = PickUpAddress()

    # XXX: Not sure exactly what this is
    parcel = fields.Int('T8904')
    quantity = fields.Int('T8905')

    gls_contract = fields.Char('T8914', length=10)
    gls_customer_id = fields.Long('T8915', length=10)


class Shipment(ShipmentMixin, CommonMixin):
    """
    Return an instance of the shipment class
    """
    def __init__(self, client):

        self.client = client

        # Initialize the values holder
        self.values = {}

        if self.client.test:
            self.special_functions = ['NOSAVE']

    def get_tags(self):
        """
        Returns all the tags (with their messages) as a list
        """
        tags = []

        def field_predicate(object):
            if object is None:
                return
            return isinstance(object, fields.BaseField)

        def field_group_predicate(object):
            if object is None:
                return
            return isinstance(object, fields.FieldGroup)

        def collect_tags(obj):
            for name, field in inspect.getmembers(
                    obj.__class__, field_predicate):
                value = getattr(obj, name)
                if value is None:
                    continue
                tags.append(':'.join((field.code, field.to_string(value))))

        # Collect the tags from fields in class
        collect_tags(self)

        # Collect the tags from fields in field groups
        for name, field_group in inspect.getmembers(
                self.__class__, field_group_predicate):
            field_group = getattr(self, name)
            collect_tags(field_group)

        return tags

    def create_label(self):
        """
        Tries to create a label on the GLS Unibox and returns a response
        """
        return self.client.request(self.get_tags())


class Response(ShipmentMixin, CommonMixin):
    """
    A response object received from GLS
    """
    def __init__(self, values):
        self.values = values

    @classmethod
    def parse(cls, string):
        """
        Parse the response.
        """
        values = {}

        start_tag_position = string.find(StartTag.code)

        # Content upto the start tag is the zpl content, followed by the tags
        zpl_content = string[:start_tag_position]
        gls_tags = string[start_tag_position:]

        for pair in gls_tags.replace(
                StartTag.code, '').replace(EndTag.code, '').split('|'):
            if ':' not in pair:
                continue
            tag, value = pair.split(':', 1)
            values[tag] = value

        values['zpl_content'] = zpl_content

        return cls(values)



# -*- coding: utf-8 -*-
"""
    fields.py

    Implemented a field descriptor
"""
import datetime
import decimal


class BaseField(object):
    def __init__(self, code, length=None):
        self.code = code
        self.length = length

    def __get__(self, obj, type=None):
        if type is None or obj is None:
            return self
        return self._from_string(
            obj.values.get(self.code)
        )

    def __set__(self, obj, value):
        if self.length is not None and \
                len(unicode(value)) > self.length:
            raise ValueError(
                'Length of %s cannot be more than %d' % (
                    self.code, self.length
                )
            )
        obj.values[self.code] = self._to_string(value)

    def __delete__(self, obj):
        if not self.mutable:
            raise RuntimeError('You cannot delete this field')
        obj.values.pop(self.code, None)

    def _from_string(self, raw_value):
        if raw_value is not None:
            return self.from_string(raw_value)

    def from_string(self, raw_value):
        return unicode(raw_value.decode('ISO-8859-1'))

    def _to_string(self, value):
        if value is not None:
            return self.to_string(value)

    def to_string(self, value):
        return unicode(value).encode('ISO-8859-1')


class Char(BaseField):
    pass


class Int(BaseField):

    def from_string(self, raw_value):
        return int(raw_value.decode('ISO-8859-1'))


class Long(BaseField):

    def from_string(self, raw_value):
        return long(raw_value.decode('ISO-8859-1'))


class Decimal(BaseField):
    def __init__(self, code, digits=(5, 1), **kwargs):
        self.digits = digits
        super(Decimal, self).__init__(code, **kwargs)

    def from_string(self, raw_value):
        return decimal.Decimal(
            raw_value.decode('ISO-8859-1').replace(',', '.')
        )

    def to_string(self, value):
        """
        Returns string representation of the field that can be sent over wire
        """
        return str(value).replace('.', ',')


class Date(BaseField):
    def __init__(self, code, separator='', **kwargs):
        """
        :param separator: Date separator
        """
        self.separator = separator
        super(Date, self).__init__(code, **kwargs)

    @property
    def format(self):
        return self.separator.join(('%d', '%m', '%Y'))

    def from_string(self, value):
        return datetime.datetime.strptime(value, self.format).date()

    def to_string(self, value):
        return value.strftime(self.format)


class Time(BaseField):

    def __init__(self, code, separator=':', **kwargs):
        """
        :param separator: Time separator
        """
        self.separator = separator
        super(Time, self).__init__(code, **kwargs)

    @property
    def format(self):
        return self.separator.join(('%H', '%M'))

    def from_string(self, value):
        return datetime.datetime.strptime(value, self.format).time()

    def to_string(self, value):
        return value.strftime(self.format)


class MultipleChoice(BaseField):

    def __init__(self, code, separator=';', **kwargs):
        """
        :param separator: separator that separates the parts
        """
        self.separator = separator
        super(MultipleChoice, self).__init__(code, **kwargs)

    @property
    def format(self):
        return self.separator.join(self.choices)

    def from_string(self, value):
        return value.split(self.separator)

    def to_string(self, value):
        return self.separator.join(value).encode('ISO-8859-1')


class FieldGroup(object):
    values = None

    def __get__(self, obj, type=None):
        if self.values is None and obj.values:
            self.values = obj.values
        return self



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import sys
from setuptools import setup
from setuptools.command.test import test as TestCommand


with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read().replace('.. :changelog:', '')

requirements = [
    # TODO: put package requirements here
]


class PyTest(TestCommand):
    user_options = []

    def initialize_options(self):
        TestCommand.initialize_options(self)
        self.pytest_args = []

    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = []
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.pytest_args)
        sys.exit(errno)


setup(
    name='gls_unibox_api',
    version='0.0.1.2',
    description="Python API for GLS Unibox Logistics System",
    long_description=readme + '\n\n' + history,
    author="Sharoon Thomas",
    author_email='info@openlabs.co.in',
    url='https://github.com/openlabs/gls_unibox_api',
    packages=[
        'gls_unibox_api',
    ],
    package_dir={'gls_unibox_api':
                 'gls_unibox_api'},
    include_package_data=True,
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    keywords='gls_unibox_api',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        "Programming Language :: Python :: 2",
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],
    tests_require=['pytest'],
    cmdclass={'test': PyTest},
)
